unit hash;

interface

uses param;

type
  TTrans=record
    key:int64;
    depth:byte;
    age:byte;
    score:integer;
    move:integer;
    flag:byte;
    mate:boolean;
  end;
  TPTrans=record
    key:integer;
    pawns:byte;
    score:integer;
  end;

var HashSize:longint;
    PHashSize:longint;
    HashSizeMB:longint;
    PHashSizeMB:longint;
    transtable:array of TTrans;
    ptranstable:array of TPTrans;
    ZorbistTable:array[1..12,0..127] of int64;
    ZorbistCast:array[1..4] of int64;
    ZorbistEnp:array[0..255] of int64;
    Zorbist32:array[1..2,0..127] of integer;
    ZorbistColor:int64;

procedure inithash;
procedure clearhash;
procedure fillzorbist;
function GetRandom:int64;
function GetRandom32:integer;
function WhatPiese(sq:byte):byte;
function GetPosHash(ply:byte):int64;
function GetPawnHash:integer;
procedure SaveToHash(depth,ply:byte;score,move:integer);
procedure LoadFromHash(ply:byte);

implementation

function GetRandom32:integer;
var res:integer;
    temp1,temp2:integer;
begin
  temp1:=Random(65536);
  temp2:=Random(65536);
  res:=(temp1 shl 16) or temp2;
  result:=res;
end;


function GetRandom:int64;
var res:int64;
    r1,r2,r3,r4:int64;
begin
  r1:=Random(65536);
  r2:=Random(65536);
  r3:=Random(65536);
  r4:=Random(65536);
  r1:=(r1 shl 16) or r4;
  r2:=(r3 shl 16) or r2;
  res:=(r1 shl 32) or r2;
  result:=res;
end;


procedure inithash;
begin
  PHashSizeMB:=HashSizeMB div 16;
  if PHashSizeMB=0 then PHashSizeMB:=1;
  HashSizeMB:=HashSizeMB-PHashSizeMB;
  HashSize:=trunc(HashSizeMB*1024*1024/(sizeof(TTrans)));
  PHashSize:=trunc(PHashSizeMB*1024*1024/(sizeof(TPTrans)));
  SetLength(TransTable,HashSize+1);
  SetLength(PTransTable,PHashSize+1);
  ClearHash;
  age:=0;
end;

procedure clearhash;
var i:longint;
begin
  for i:=1 to HashSize do
  begin
    TransTable[i].key:=0;
    TransTable[i].age:=0;
    TransTable[i].depth:=0;
    TransTable[i].score:=0;
    TransTable[i].move:=0;
    TransTable[i].flag:=0;
    TransTable[i].mate:=false;
  end;
  for i:=1 to PHashSize do
  begin
    pTransTable[i].key:=0;
    pTransTable[i].pawns:=0;
    pTransTable[i].score:=0;
  end;
  age:=0;
end;

procedure fillzorbist;
var i,j:integer;
begin
  for i:=1 to 12 do
    for j:=0 to 127 do
      ZorbistTable[i,j]:=GetRandom;
  for i:=1 to 4 do
    ZorbistCast[i]:=GetRandom;
  for i:=0 to 255 do
    ZorbistEnp[i]:=GetRandom;
  for i:=1 to 2 do
    for j:=0 to 127 do
      Zorbist32[i,j]:=GetRandom32;
  ZorbistEnp[255]:=0;
  ZorbistColor:=GetRandom;
end;

function WhatPiese(sq:byte):byte;
var res:byte;
begin
  res:=board[sq] and 7;
  if (board[sq] and 128)<>0 then res:=res+6;
  result:=res;
end;

function GetPosHash(ply:byte):int64;
var res:int64;
    i,p:integer;
begin
  res:=0;
  for i:=0 to 127 do
    if (i and 136)=0 then
    begin
      p:=WhatPiese(i);
      if p<>0 then res:=res xor ZorbistTable[p,i];
    end;
  if (tree[ply].castles and whiteking)<>0 then res:=res xor ZorbistCast[1];
  if (tree[ply].castles and whitequeen)<>0 then res:=res xor ZorbistCast[2];
  if (tree[ply].castles and blackking)<>0 then res:=res xor ZorbistCast[3];
  if (tree[ply].castles and blackqueen)<>0 then res:=res xor ZorbistCast[4];
  res:=res xor ZorbistEnp[tree[ply].enpass];
  if tree[ply].tomove=1 then res:=res xor ZorbistColor;
  result:=res;
end;

function GetPawnHash:integer;
var i,p,res:integer;
begin
  res:=0;
  for i:=0 to 127 do
    if (i and 136)=0 then
    begin
      p:=WhatPiese(i);
      if (p=6) or (p=12) then
      begin
        if p=6 then p:=1;
        if p=12 then p:=2;
        res:=res xor Zorbist32[p,i];
      end;
    end;
  result:=res;
end;


procedure SaveToHash(depth,ply:byte;score,move:integer);
var ind:longint;
begin
  if abs(score)+100>inf then exit;
  ind:=abs(tree[ply].boardkey) mod HashSize;
  if (depth>=TransTable[ind].depth) or (age<>TransTable[ind].age) then
  begin
    TransTable[ind].key:=tree[ply].boardkey;
    TransTable[ind].depth:=depth;
    TransTable[ind].score:=score;
    TransTable[ind].move:=move;
    TransTable[ind].age:=age;
    TransTable[ind].flag:=tree[ply].hflag;
    TransTable[ind].mate:=tree[ply].hmate;
  end;
end;

procedure LoadFromHash(ply:byte);
var ind:longint;
begin
  ind:=abs(tree[ply].boardkey) mod HashSize;
  tree[ply].hflag:=0;
  tree[ply].hmove:=0;
  tree[ply].hmate:=false;
  if TransTable[ind].key=tree[ply].boardkey then
  begin
    tree[ply].hmove:=TransTable[ind].move;
    tree[ply].hflag:=TransTable[ind].flag;
    tree[ply].hscore:=TransTable[ind].score;
    tree[ply].hdepth:=TransTable[ind].depth;
    tree[ply].hmate:=TransTable[ind].mate;
  end;
end;

end.
